@@ -144,6 +144,36 @@ static void fts_backend_xapian_clean(icu::UnicodeString *t)
 	fts_backend_xapian_trim(t);
 }
 
+static long fts_backend_xapian_clean_header(const char * hdr)
+{
+	if(hdr == NULL) return -1;
+	long l = strlen(hdr);
+	if(l>=200) return -1;
+
+	char h[200];
+
+	long i=0,j=0;
+        while(j<l)
+        {
+        	if((hdr[j]>' ') && (hdr[j]!='"') && (hdr[j]!='\'') && (hdr[j]!='-'))
+                {
+                        h[i]=std::tolower(hdr[j]);
+                        i++;
+                }
+                j++;
+        }       
+        h[i]=0;
+
+        i=0;
+        while((i<HDRS_NB) && (strcmp(h,hdrs_emails[i])!=0)) i++;
+                        
+	if(i>=HDRS_NB) i=-1;
+
+	if(i==HDRS_NB-1) i=HDR_BODY;
+
+	return i;
+}
+
 static void fts_backend_xapian_get_lock(struct xapian_fts_backend *backend, long verbose, const char *s)
 {
 	std::unique_lock<std::timed_mutex> *lck;
@@ -152,7 +182,7 @@ static void fts_backend_xapian_get_lock(struct xapian_fts_backend *backend, long
 #pragma GCC diagnostic ignored "-Wdeprecated-declarations"
         while(!(lck->try_lock_for(std::chrono::milliseconds(1000 + std::rand() % 1000))))
 	{
-        	if(verbose>1) 
+        	if(verbose>0) 
 		{
 			std::string sl("FTS Xapian: Waiting unlock... (");
 			sl.append(s);
@@ -161,7 +191,7 @@ static void fts_backend_xapian_get_lock(struct xapian_fts_backend *backend, long
                 }
 	}
 #pragma GCC diagnostic pop
-	if(verbose>1)
+	if(verbose>0)
 	{
 		std::string sl("FTS Xapian: Got lock (");
 		sl.append(s);
@@ -234,6 +264,16 @@ static bool fts_backend_xapian_sqlite3_dict_open(struct xapian_fts_backend *back
 		return FALSE;
         }
 
+	zErrMsg =0;
+	if(sqlite3_exec(backend->ddb,createDictIndexes,NULL,0,&zErrMsg) != SQLITE_OK )
+        {
+                i_error("FTS Xapian: Can not execute (%s) : %s",createDictIndexes,zErrMsg);
+                if(zErrMsg!=NULL) sqlite3_free(zErrMsg);
+                sqlite3_close(backend->ddb);
+                backend->ddb = NULL;
+                return FALSE;
+        }
+
 	zErrMsg =0;
 	if(sqlite3_exec(backend->ddb,createTmpTable,NULL,0,&zErrMsg) != SQLITE_OK )
         {
@@ -246,17 +286,17 @@ static bool fts_backend_xapian_sqlite3_dict_open(struct xapian_fts_backend *back
 	return TRUE;
 }
 
-static int fts_backend_xapian_sqlite3_dict_add(struct xapian_fts_backend *backend, icu::UnicodeString *t)
+static int fts_backend_xapian_sqlite3_dict_add(struct xapian_fts_backend *backend, long h, icu::UnicodeString *t)
 {
 	std::string sql;
         sql.clear();
         t->toUTF8String(sql);
-        sql=replaceTmpWord + sql + "'," + std::to_string(sql.length()) + ")";
-                
-	char * zErrMsg =0;
+        sql=replaceTmpWord + sql + "', " + std::to_string(h) + ", " + std::to_string(sql.length()) + ");";
+
+	char * zErrMsg = 0;
 	if(sqlite3_exec(backend->ddb,sql.c_str(),NULL,0,&zErrMsg) != SQLITE_OK )
         {
-		syslog(LOG_ERR,"FTS Xapian: Can not replace keyword : %s",sql.c_str(),zErrMsg);
+		syslog(LOG_ERR,"FTS Xapian: Can not replace keyword (%s) : %s",sql.c_str(),zErrMsg);
                 if(zErrMsg!=NULL) sqlite3_free(zErrMsg);
 		return 1;
 	}
@@ -309,7 +349,7 @@ class XResultSet
 class XQuerySet
 {
 	private:
-		const char * header;
+		long header;
 		icu::UnicodeString * text;
 		XQuerySet ** qs;
 		Xapian::Query::op global_op;
@@ -322,8 +362,8 @@ class XQuerySet
 	XQuerySet()
 	{
 		qsize=0; qs=NULL;
-		limit=2;
-		header=NULL;
+		limit=1;
+		header=-1;
 		text=NULL;
 		global_op = Xapian::Query::op::OP_OR;
 	}
@@ -332,8 +372,8 @@ class XQuerySet
 	{
 		qsize=0; qs=NULL;
 		limit=2;
-		if(l>2) { limit=l; }
-		header=NULL;
+		if(l>limit) { limit=l; }
+		header=-1;
 		text=NULL;
 		global_op=op;
 	}
@@ -358,26 +398,19 @@ class XQuerySet
 	{
 		std::string s = std::to_string(uid);
 		icu::UnicodeString t(s.c_str());
-		add(hdrs_emails[0],&t,false);
+		add(0,&t,false);
 	}
 		
-	void add(const char * h2, icu::UnicodeString *t, bool is_neg)
+	void add(long h, icu::UnicodeString *t, bool is_neg)
 	{
-		if(h2==NULL) return;
 		if(t==NULL) return;
+		fts_backend_xapian_clean(t);
+                if(t->length()<limit) return;
 
-		icu::UnicodeString h(h2);
-		h.trim();
-		h.toLower();
-                if(h.length()<1) return;
-
-		long i,j,k;
+		long i,j;
 		XQuerySet * q2;
 		icu::UnicodeString *r;
 
-		fts_backend_xapian_clean(t);
-		if(t->length()<limit) return;
-	
 		i = t->lastIndexOf(CHAR_SPACE);
                 if(i>0)
                 {
@@ -393,18 +426,18 @@ class XQuerySet
                         {
                                 j = t->length();
                                 r = new icu::UnicodeString(*t,i+1,j-i-1);
-                                q2->add(h2,r,false);
+                                q2->add(h,r,false);
                                 delete(r);
                                 t->truncate(i);
                                 fts_backend_xapian_trim(t);
                                 i = t->lastIndexOf(CHAR_SPACE);
                         }
-                        q2->add(h2,t,false);
+                        q2->add(h,t,false);
                         if(q2->count()>0) add(q2); else delete(q2);
                         return;
                 }
 
-		if(h.compare(XAPIAN_WILDCARD)==0)
+		if(h<0)
 		{
 			if(is_neg)
 			{
@@ -414,37 +447,24 @@ class XQuerySet
 			{
                         	q2 = new XQuerySet(Xapian::Query::OP_OR,limit);
 			}
-			for(i=1;i<HDRS_NB;i++)
+			for(i=1;i<HDRS_NB-1;i++)
 			{
-				if(i!=XAPIAN_EXPUNGE_HEADER) q2->add(hdrs_emails[i],t,false);
+				q2->add(i,t,false);
 			}
 			add(q2);
 			return;
 		}
-                else
-		{
-                        i=0;
-                        while((i<HDRS_NB) && (h.compare(hdrs_emails[i])!=0))
-                        {
-                                i++;
-                        }
-                        if(i>=HDRS_NB)
-                        {
-                                return;
-                        }
-			h2=hdrs_emails[i];
-                }
-
+		
 		if(text==NULL)
 		{
 			text=new icu::UnicodeString(*t);
-			header=h2;
+			header=h;
 			item_neg=is_neg;
 			return;
 		}
 
                 q2 = new XQuerySet(Xapian::Query::OP_AND,limit);
-		q2->add(h2,t,is_neg);
+		q2->add(h,t,is_neg);
 		add(q2);
 	}
 
@@ -479,9 +499,8 @@ class XQuerySet
 		if(text!=NULL)
 		{
 			if(item_neg) s.append("NOT ( ");
-			s.append(header);
-			s.append(":");
-			s.append("\"");
+			s.append(hdrs_emails[header]);
+			s.append(":\"");
 			text->toUTF8String(s);
 			s.append("\"");
 			if(item_neg) s.append(")");
@@ -521,13 +540,13 @@ class XQuerySet
 
 		if(text!=NULL)
                 {
-			std::string s(header);
+			std::string s(hdrs_query[header]);
                         s.append(":\"");
                         text->toUTF8String(s);
                         s.append("\"");
 
 			Xapian::QueryParser * qp = new Xapian::QueryParser();
-			for(int i=0; i< HDRS_NB; i++) qp->add_prefix(hdrs_emails[i], hdrs_xapian[i]);
+			for(int i=0; i< HDRS_NB-1; i++) qp->add_prefix(hdrs_query[i], hdrs_xapian[i]);
 			qp->set_database(*db);
 			q = new Xapian::Query(qp->parse_query(s.c_str(),Xapian::QueryParser::FLAG_DEFAULT));
 			delete (qp);
@@ -571,7 +590,7 @@ class XDoc
 	private:
                 std::vector<icu::UnicodeString *> * terms;
 		std::vector<icu::UnicodeString *> * strings;
-		std::vector<const char *> * headers;
+		std::vector<long> * headers;
 		struct xapian_fts_backend *backend;
 
 	public:
@@ -593,7 +612,7 @@ class XDoc
 
 		strings = new std::vector<icu::UnicodeString *>;
 		strings->clear();
-		headers = new std::vector<const char *>;
+		headers = new std::vector<long>;
 		headers->clear();
 		terms = new std::vector<icu::UnicodeString *>;
 		terms->clear();
@@ -637,7 +656,7 @@ class XDoc
 		return s;
 	}
 
-	void raw_load(const char *h, const char *d, int32_t size, long verbose, const char * title)
+	void raw_load(long h, const char *d, int32_t size, long verbose, const char * title)
 	{
 		icu::UnicodeString * t;
                 {
@@ -671,20 +690,22 @@ class XDoc
                 return terms_add(w,pos,n);
         }
 
-	void terms_push(const char *h, icu::UnicodeString *t)
+	void terms_push(long h, icu::UnicodeString *t)
 	{
 		fts_backend_xapian_trim(t);
-		long n = t->length();
-		
+		unsigned long n = t->length();
+		long m = XAPIAN_TERM_SIZELIMIT - strlen(hdrs_xapian[h]) - 1;
+	
 		if(n>=fts_xapian_settings.partial)
 		{
-			t->truncate(XAPIAN_TERM_SIZELIMIT-strlen(h)-1);	
-			while(fts_backend_xapian_icutochar_length(t)>=XAPIAN_TERM_SIZELIMIT-strlen(h)-1)
+			t->truncate(m);	
+			while(fts_backend_xapian_icutochar_length(t)>=m)
 			{
 				t->truncate(t->length()-1);
 			}
-			fts_backend_xapian_sqlite3_dict_add(backend,t);
-			t->insert(0,h);
+			fts_backend_xapian_sqlite3_dict_add(backend,h,t);
+			ndict++;
+			t->insert(0,hdrs_xapian[h]);
 			terms_add(t,0,terms->size());
                 }
                 delete(t);
@@ -693,7 +714,7 @@ class XDoc
 	bool terms_create(long verbose, const char * title)
 	{
 		icu::UnicodeString *t;
-		const char * h;
+		long h;
 		long k;
 		
 		while((terms->size()<XAPIAN_MAXTERMS_PERDOC) && (strings->size()>0))
@@ -789,11 +810,11 @@ class XDocsWriter
 		}
                 catch(Xapian::DatabaseLockError e)
                 {
-                        syslog(LOG_WARNING,"%sCan't lock the DB : %s - %s",title,e.get_type(),e.get_msg());
+                        syslog(LOG_WARNING,"%sCan't lock the DB : %s - %s",title,e.get_type(),e.get_msg().c_str());
 		}
                 catch(Xapian::Error e)
                 {
-			syslog(LOG_WARNING,"%sCan't open the DB RW : %s - %s",title,e.get_type(),e.get_msg());
+			syslog(LOG_WARNING,"%sCan't open the DB RW : %s - %s",title,e.get_type(),e.get_msg().c_str());
                 }
 		return false;
 	}
@@ -838,7 +859,7 @@ class XDocsWriter
 		{
 			t = new std::thread(fts_backend_xapian_worker,this);
 		}
-		catch(std::exception e)
+		catch(std::exception const& e)
 		{
 			std::string s(title);
 			s.append("Thread error ");
@@ -873,17 +894,25 @@ class XDocsWriter
 			{
 				try
                         	{
-					syslog(LOG_WARNING,"%sCommitting %ld docs due to low free memory (%ld MB vs %ld MB) or Cached docs exceeded (%ld vs %ld limit)",title,backend->pending,(long)(m/1024.0f),lowmemory,backend->pending,XAPIAN_WRITING_CACHE);
+					if(backend->pending > XAPIAN_WRITING_CACHE) 
+					{
+						syslog(LOG_WARNING,"%sCommitting %ld docs due to cached docs exceeded (%ld vs %ld limit)",title,backend->pending,backend->pending,XAPIAN_WRITING_CACHE);
+					}
+					else 
+					{
+						syslog(LOG_WARNING,"%sCommitting %ld docs due to low free memory (%ld MB vs %ld MB)",title,backend->pending,(long)(m/1024.0f),lowmemory);
+					}
                         	        backend->dbw->close();
                         	        delete(backend->dbw);
-                        	        backend->dbw=NULL;
+					if(verbose>0) syslog(LOG_INFO,"%sClosed Xapian DB %s",title,backend->xap_db);
+                        	        backend->dbw = NULL;
                         	        backend->pending = 0;
                         	}
                         	catch(Xapian::Error e)
                         	{
-					syslog(LOG_ERR,"%sCan't commit DB1 : %s - %s",title,e.get_type(),e.get_msg());
+					syslog(LOG_ERR,"%sCan't commit DB1 : %s - %s",title,e.get_type(),e.get_msg().c_str());
                         	}
-                        	catch(std::exception e)
+                        	catch(std::exception const& e)
                         	{
 					syslog(LOG_ERR,"%sCan't commit DB2 : %s",title,e.what());
                         	}
@@ -974,7 +1003,7 @@ class XDocsWriter
 				if(verbose>0) syslog(LOG_INFO,"%sPushing : %s",title,doc->getDocSummary().c_str());
                         	if(doc->nterms > 0)
                         	{
-					long m = checkMemory();
+					checkMemory();
 					fts_backend_xapian_get_lock(backend, verbose, title);
 					if(checkDB())
 					{
@@ -990,9 +1019,9 @@ class XDocsWriter
                         	        	}
 						catch(Xapian::Error e)
                                	                {
-                               	                 	syslog(LOG_ERR,"%sCan't write doc1 %s : %s - %s",title,doc->getDocSummary().c_str(),e.get_type(),e.get_msg());
+                               	                 	syslog(LOG_ERR,"%sCan't write doc1 %s : %s - %s",title,doc->getDocSummary().c_str(),e.get_type(),e.get_msg().c_str());
                                	                }
-                               	                catch(std::exception e)
+                               	                catch(std::exception const & e)
                                	                {
                                	                        syslog(LOG_ERR,"%sCan't write doc2 %s : %s",title,doc->getDocSummary().c_str(),e.what());
                                	                }
@@ -1084,7 +1113,7 @@ static void fts_backend_xapian_close_db(Xapian::WritableDatabase * dbw,const cha
         {
                 syslog(LOG_ERR, "FTS Xapian: Can't close Xapian DB (%s) %s : %s - %s %s",boxname,dbpath,e.get_type(),e.get_msg().c_str(),e.get_error_string());
         }
-	catch(std::exception e)
+	catch(std::exception const& e)
         {
                 syslog(LOG_ERR, "FTS Xapian : Closing db (%s) error %s",dbpath,e.what());
         }
@@ -1166,15 +1195,6 @@ static void fts_backend_xapian_close(struct xapian_fts_backend *backend, const c
 	}
 }
 
-static bool fts_backend_xapian_isnormalprocess()
-{
-    	char buff[PATH_MAX];
-    	buff[0]=0;
-    	ssize_t len = ::readlink("/proc/self/exe", buff, sizeof(buff)-1);
-    	if (len>0)  buff[len] = 0;
-    	return (strstr(buff,"doveadm")==NULL);
-}
-
 XResultSet * fts_backend_xapian_query(Xapian::Database * dbx, XQuerySet * query, long limit=0)
 {
 	if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: fts_backend_xapian_query (%s)",query->get_string().c_str());
@@ -1216,8 +1236,6 @@ static int fts_backend_xapian_unset_box(struct xapian_fts_backend *backend)
 {
 	if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: Unset box '%s' (%s)",backend->boxname,backend->guid);
 
-	long commit_time = fts_backend_xapian_current_time();
-
 	fts_backend_xapian_close(backend,"unset box");
 	fts_backend_xapian_oldbox(backend);
 
@@ -1236,7 +1254,10 @@ static int fts_backend_xapian_unset_box(struct xapian_fts_backend *backend)
 		backend->exp_db = NULL;
 
 		i_free(backend->dict_db);
-                backend->exp_db = NULL;
+                backend->dict_db = NULL;
+
+		i_free(backend->version_file);
+		backend->version_file = NULL;
 	}
 
 	return 0;
@@ -1310,32 +1331,65 @@ static int fts_backend_xapian_set_box(struct xapian_fts_backend *backend, struct
 	backend->xap_db = i_strdup_printf("%s/db_%s",backend->path,mb);
 	backend->exp_db = i_strdup_printf("%s%s",backend->xap_db,suffixExp);
 	backend->dict_db = i_strdup_printf("%s%s",backend->xap_db,suffixDict);
+	backend->version_file = i_strdup_printf("%s_v%s",backend->xap_db,XAPIAN_PLUGIN_VERSION);
 
 	struct stat sb;
+	// Existence of current version
+	if(!( (stat(backend->version_file, &sb)==0) && S_ISREG(sb.st_mode)))
+        {
+		i_info("FTS Xapian: New version of the plugin : %s for %s",XAPIAN_PLUGIN_VERSION,backend->boxname);	
+
+		// Deleting existing indexes
+                std::filesystem::remove_all(backend->xap_db);
+		for(auto& f : std::filesystem::directory_iterator(backend->path)) 
+		{
+			if((f.is_regular_file()) && (f.path().string().find(backend->xap_db) == 0))
+			{
+				if(fts_xapian_settings.verbose>0) i_warning("FTS Xapian: Deleting %s",f.path().c_str());
+				std::filesystem::remove(f.path());
+			}
+		}
+
+		// Write new version file                
+		FILE * f = fopen(backend->version_file,"w+");
+		fprintf(f,"%s",XAPIAN_PLUGIN_VERSION);
+		fclose(f);
+	}
+
 	// Verify existence of Dict db
 	if(!( (stat(backend->dict_db, &sb)==0) && S_ISREG(sb.st_mode)))
 	{
 		i_warning("FTS Xapian: '%s' (%s) dictionnary does not exist. Creating it",backend->boxname,backend->dict_db);
-                // Deleting existing indexes
-                try
-                {
-                        std::filesystem::remove_all(backend->xap_db);
-                        std::filesystem::remove(backend->exp_db);
-                }
-                catch(std::exception e)
-                {
-                        i_error("FTS Xapian: Can not delete old files %s",e.what());
-                }
-
-		// Creating dictionnary
 		if(fts_backend_xapian_sqlite3_dict_open(backend)) sqlite3_close(backend->ddb);
 		backend->ddb = NULL;
+		std::filesystem::remove_all(backend->xap_db);
 	}
 	
+	// Verify existence of Xapian db
+	{
+		char * t = i_strdup_printf("%s/termlist.glass",backend->xap_db);
+        	if(!( (stat(t, &sb)==0) && S_ISREG(sb.st_mode)))
+        	{
+			std::filesystem::remove(backend->exp_db);
+                	i_info("FTS Xapian: '%s' (%s) indexes do not exist. Initializing DB",backend->boxname,backend->xap_db);
+                	try
+                	{
+                	        Xapian::WritableDatabase * db = new Xapian::WritableDatabase(backend->xap_db,Xapian::DB_CREATE_OR_OVERWRITE | Xapian::DB_BACKEND_GLASS);
+                	        db->close();
+                	        delete(db);
+                	}
+                	catch(Xapian::Error e)
+                	{
+                	        i_error("FTS Xapian: Can't create Xapian DB (%s) %s : %s - %s %s",backend->boxname,backend->xap_db,e.get_type(),e.get_msg().c_str(),e.get_error_string());
+                	}
+        	}
+        	i_free(t);
+	}
+
         // Verify existence of Exp db
         if(!( (stat(backend->exp_db, &sb)==0) && S_ISREG(sb.st_mode)))
-	{
-		i_warning("FTS Xapian: '%s' (%s) expunge does not exist. Creating it",backend->boxname,backend->exp_db);
+        {
+                i_warning("FTS Xapian: '%s' (%s) expunge does not exist. Creating it",backend->boxname,backend->exp_db);
                 sqlite3 * db = NULL;
                 if(sqlite3_open_v2(backend->exp_db,&db,SQLITE_OPEN_FULLMUTEX | SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE,NULL) != SQLITE_OK )
                 {
@@ -1353,25 +1407,6 @@ static int fts_backend_xapian_set_box(struct xapian_fts_backend *backend, struct
                 }
         }
 
-	// Verify existence of Xapian db
-	{
-		char * t = i_strdup_printf("%s/termlist.glass",backend->xap_db);
-        	if(!( (stat(t, &sb)==0) && S_ISREG(sb.st_mode)))
-        	{
-                	i_info("FTS Xapian: '%s' (%s) indexes do not exist. Initializing DB",backend->boxname,backend->xap_db);
-                	try
-                	{
-                	        Xapian::WritableDatabase * db = new Xapian::WritableDatabase(backend->xap_db,Xapian::DB_CREATE_OR_OVERWRITE | Xapian::DB_BACKEND_GLASS);
-                	        db->close();
-                	        delete(db);
-                	}
-                	catch(Xapian::Error e)
-                	{
-                	        i_error("FTS Xapian: Can't create Xapian DB (%s) %s : %s - %s %s",backend->boxname,backend->xap_db,e.get_type(),e.get_msg().c_str(),e.get_error_string());
-                	}
-        	}
-        	i_free(t);
-	}
         backend->threads.clear();
 	backend->total_docs =0;
 
@@ -1380,7 +1415,7 @@ static int fts_backend_xapian_set_box(struct xapian_fts_backend *backend, struct
 
 static void fts_backend_xapian_build_qs(XQuerySet * qs, struct mail_search_arg *a, const char * dict=NULL)
 {
-	const char * hdr;
+	long hdr;
 
 	if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: fts_backend_xapian_build_qs");
 
@@ -1388,29 +1423,21 @@ static void fts_backend_xapian_build_qs(XQuerySet * qs, struct mail_search_arg *
 	{
 		switch (a->type)
 		{
-			case SEARCH_TEXT:
-			case SEARCH_BODY:
+			case SEARCH_TEXT: hdr = -1; break;
+			case SEARCH_BODY: hdr = 8; break;
 			case SEARCH_HEADER:
 			case SEARCH_HEADER_ADDRESS:
-			case SEARCH_HEADER_COMPRESS_LWSP: break;
+			case SEARCH_HEADER_COMPRESS_LWSP: 
+				if((a->hdr_field_name == NULL)||(strlen(a->hdr_field_name)<1))
+				{
+					hdr = -1;
+					break;
+				}
+				hdr=fts_backend_xapian_clean_header(a->hdr_field_name); 
+				if(hdr >= 0) break;
 			default: a = a->next; continue;
 		}
 
-		if((a->hdr_field_name == NULL)||(strlen(a->hdr_field_name)<1))
-		{
-			if(a->type == SEARCH_BODY)
-			{
-				hdr="body";
-			}
-			else
-			{
-				hdr=XAPIAN_WILDCARD;
-			}
-		}
-		else
-		{
-			hdr=a->hdr_field_name;
-		}
 		if((a->value.str == NULL) || (strlen(a->value.str)<1))
 		{
 			XQuerySet * q2;
@@ -1434,35 +1461,26 @@ static void fts_backend_xapian_build_qs(XQuerySet * qs, struct mail_search_arg *
 		}
 		else if(dict != NULL)
 		{
-			//Clean header
-			std::string header; header.clear();
-			long i=0,j=strlen(hdr);
-                        while(i<j)
-                        {
-                        	if((hdr[i]>' ') && (hdr[i]!='"') && (hdr[i]!='\'') && (hdr[i]!='-'))
-                                {
-                                        header+=tolower(hdr[i]);
-                                }
-                                i++;
-			}
-
 			// Find key words
 			icu::StringPiece sp(a->value.str);
                         icu::UnicodeString t = icu::UnicodeString::fromUTF8(sp);
                         fts_backend_xapian_clean(&t);
-			i = t.lastIndexOf(CHAR_SPACE),j;
+			long j, i = t.lastIndexOf(CHAR_SPACE);
+			unsigned long l;
 			icu::UnicodeString *k;
 			std::vector<icu::UnicodeString *> keys; keys.clear();
 			while(i>0)
 			{
 				j = t.length();
                                 k = new icu::UnicodeString(t,i+1,j-i-1);
-				if(k->length()>1) { keys.push_back(k); } else delete(k);
+				l = k->length();
+				if(l >= fts_xapian_settings.partial) { keys.push_back(k); } else delete(k);
 				t.truncate(i);
                                	fts_backend_xapian_trim(&t);
                                	i = t.lastIndexOf(CHAR_SPACE);
                         }
-			if(t.length()>1) 
+			l = t.length();
+			if(l>=fts_xapian_settings.partial) 
 			{
 				keys.push_back(new icu::UnicodeString (t));
 			}
@@ -1475,6 +1493,7 @@ static void fts_backend_xapian_build_qs(XQuerySet * qs, struct mail_search_arg *
                                 syslog(LOG_ERR,"FTS Xapian: Can not open %s : %s",dict,sqlite3_errmsg(db));
                                 return;
                         }
+
 			// Generate query
 			XQuerySet * q1, *q2;
                         if(a->match_not)
@@ -1490,20 +1509,29 @@ static void fts_backend_xapian_build_qs(XQuerySet * qs, struct mail_search_arg *
 				std::vector<icu::UnicodeString *> st; st.clear();
 				std::string sql=searchDict1;
 				ki->toUTF8String(sql);
+				if(hdr<0)
+				{
+					sql+="%'";
+				}
+				else
+				{
+					sql+="%' and header=" + std::to_string(hdr);
+				}
 				sql +=searchDict2;
+				
 				zErrMsg =0;
 				if(sqlite3_exec(db,sql.c_str(),fts_backend_xapian_sqlite3_vector_icu,&st,&zErrMsg) != SQLITE_OK )
                                 {
-                                        syslog(LOG_ERR,"FTS Xapian: Can not search keyword : %s",sql.c_str(),zErrMsg);
+                                        syslog(LOG_ERR,"FTS Xapian: Can not search keyword (%s) : %s",sql.c_str(),zErrMsg);
                                         if(zErrMsg!=NULL) sqlite3_free(zErrMsg);
                                 }
 				q2 = new XQuerySet(Xapian::Query::OP_OR,qs->limit);
                                 for(auto &term : st)
                                 {
-                                        q2->add(header.c_str(),term,false);
+                                        q2->add(hdr,term,false);
                                         delete(term);
                                 }
-                                q1->add(q2);
+                                if(q2->count()>0) { q1->add(q2); } else { delete(q2); }
 				delete(ki);
 			}
 			qs->add(q1);
@@ -1511,21 +1539,11 @@ static void fts_backend_xapian_build_qs(XQuerySet * qs, struct mail_search_arg *
 		}
 		else
 		{
-			long i=0,j=strlen(hdr);
-			std::string f2; f2.clear();
-			while(i<j)
-			{
-				if((hdr[i]>' ') && (hdr[i]!='"') && (hdr[i]!='\'') && (hdr[i]!='-'))
-				{
-					f2+=tolower(hdr[i]);
-				}
-				i++;
-			}
 			icu::StringPiece sp(a->value.str);
 			icu::UnicodeString t = icu::UnicodeString::fromUTF8(sp);
 			fts_backend_xapian_clean(&t);
 	
-			qs->add(f2.c_str(),&t,a->match_not);
+			qs->add(hdr,&t,a->match_not);
 		}
 		a->match_always=true;
 		a = a->next;
