@@ -11,6 +11,7 @@
 extern "C" {
 #include "fts-xapian-plugin.h"
 }
+#include "fts-backend-xapian.h"
 #include <dirent.h>
 #include <filesystem>
 #include <unicode/utypes.h>
@@ -40,6 +41,7 @@ struct xapian_fts_backend
 
 	char * xap_db;
 	char * exp_db;
+	char * version_file;
 	char * dict_db;
 	long dict_nb;
 
@@ -54,7 +56,11 @@ struct xapian_fts_backend
 	std::vector<XDocsWriter *> threads;
 	std::timed_mutex mutex;
 	std::unique_lock<std::timed_mutex> * mutex_t;
-	long max_threads;
+	unsigned int max_threads;
+
+#ifdef FTS_DOVECOT24
+	struct event *event;
+#endif
 
 	long lastuid;
 	long total_docs;
@@ -72,6 +78,11 @@ struct xapian_fts_backend_update_context
 
 static struct fts_xapian_settings fts_xapian_settings;
 
+struct event_category event_category_fts_xapian = {
+        .parent = &event_category_fts,
+        .name = XAPIAN_LABEL
+};
+
 #include "fts-backend-xapian-functions.cpp"
 
 
@@ -106,24 +117,41 @@ static int fts_backend_xapian_init(struct fts_backend *_backend, const char **er
 	backend->path = NULL;
 	backend->old_guid = NULL;
 	backend->old_boxname = NULL;
-	if(fts_xapian_settings.maxthreads>0) 
-	{
-		backend->max_threads=fts_xapian_settings.maxthreads;
-	}
-	else
-	{
-		backend->max_threads = std::thread::hardware_concurrency()-1;
-	}
-	if(backend->max_threads<2) backend->max_threads = 2;
 
 	struct fts_xapian_user *fuser = FTS_XAPIAN_USER_CONTEXT(_backend->ns->user);
+	
+#ifdef FTS_DOVECOT24
+	backend->event = event_create(_backend->event);
+        event_add_category(backend->event, &event_category_fts_xapian);
+
+	if (fts_xapian_mail_user_get(_backend->ns->user, backend->event, &fuser, error_r) < 0) 
+	{
+                event_unref(&backend->event);
+                return -1;
+        }
+	fts_xapian_settings.verbose = fuser->set->verbose;
+	fts_xapian_settings.maxthreads = fuser->set->maxthreads;
+	fts_xapian_settings.partial = fuser->set->partial;
+	fts_xapian_settings.lowmemory = fuser->set->lowmemory;
+#else	
 	fts_xapian_settings = fuser->set;
+#endif
+
+	if(fts_xapian_settings.maxthreads>0)
+        {
+                backend->max_threads=fts_xapian_settings.maxthreads;
+        }
+        else
+        {
+                backend->max_threads = std::thread::hardware_concurrency()-1;
+        }
+        if(backend->max_threads<2) backend->max_threads = 2;
 
 	if(fts_backend_xapian_set_path(backend)<0) return -1;
 
 	openlog("xapian-docswriter",0,LOG_MAIL);
 
-        if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Starting version %s with partial=%ld verbose=%ld max_threads=%ld lowmemory=%ld MB", XAPIAN_PLUGIN_VERSION, fts_xapian_settings.partial,fts_xapian_settings.verbose,backend->max_threads,fts_xapian_settings.lowmemory);
+        if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Starting version %s with partial=%d verbose=%d max_threads=%u lowmemory=%d MB", XAPIAN_PLUGIN_VERSION, fts_xapian_settings.partial,fts_xapian_settings.verbose,backend->max_threads,fts_xapian_settings.lowmemory);
 
 	return 0;
 }
@@ -145,6 +173,10 @@ static void fts_backend_xapian_deinit(struct fts_backend *_backend)
 	if(backend->path != NULL) i_free(backend->path);
 	backend->path = NULL;
 
+#ifdef FTS_DOVECOT24
+	event_unref(&backend->event);
+#endif
+
 	i_free(backend);
 
 	closelog();
@@ -260,6 +292,7 @@ static bool fts_backend_xapian_update_set_build_key(struct fts_backend_update_co
 
 	ctx->tbi_isfield=false;
 	ctx->tbi_uid=0;
+	ctx->tbi_field=NULL;
 
 	if(backend->guid == NULL)
 	{
@@ -274,13 +307,10 @@ static bool fts_backend_xapian_update_set_build_key(struct fts_backend_update_co
 		backend->old_boxname = i_strdup(backend->boxname);
 	}
 
-	const char * field=key->hdr_name;
 	const char * type = key->body_content_type;
 	const char * disposition = key->body_content_disposition;
 
-	if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: New part (Header=%s,Type=%s,Disposition=%s)",field,type,disposition);
-
-	if(!fts_backend_xapian_sqlite3_dict_open(backend)) return FALSE;
+	if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: New part (Header=%s,Type=%s,Disposition=%s)",key->hdr_name,type,disposition);
 
 	// Verify content-type
 	if(key->type == FTS_BACKEND_BUILD_KEY_BODY_PART_BINARY)
@@ -295,6 +325,8 @@ static bool fts_backend_xapian_update_set_build_key(struct fts_backend_update_co
 		return FALSE;
 	}
 
+	if(!fts_backend_xapian_sqlite3_dict_open(backend)) return FALSE;
+
 	// Verify content-disposition
 	ctx->isattachment=false;
 	if((disposition != NULL) && ((strstr(disposition,"filename=")!=NULL) || (strstr(disposition,"attachment")!=NULL)))
@@ -303,35 +335,16 @@ static bool fts_backend_xapian_update_set_build_key(struct fts_backend_update_co
 		ctx->isattachment=true;
 	}
 
-	// Fill-in field
-	if(field==NULL)
+	long field = HDR_BODY;
+	if(key->hdr_name!=NULL)
 	{
-		field="body";
-	}
-
-	long i=0,j=strlen(field);
-	std::string f2;
-	while(i<j)
-	{
-		if((field[i]>' ') && (field[i]!='"') && (field[i]!='\'') && (field[i]!='-'))
+		field = fts_backend_xapian_clean_header(key->hdr_name);
+		if(field<0)
 		{
-			f2+=tolower(field[i]);
+			if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: Unknown header '%s' of part",key->hdr_name);
+			return FALSE;
 		}
-		i++;
-	}
-	ctx->tbi_field=i_strdup(f2.c_str());
-
-	i=0;
-	while((i<HDRS_NB) && (strcmp(ctx->tbi_field,hdrs_emails[i])!=0))
-	{
-		i++;
-	}
-	if(i>=HDRS_NB)
-	{
-		if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: Unknown header '%s' of part",ctx->tbi_field);
-		i_free(ctx->tbi_field);
-		ctx->tbi_field=NULL;
-		return FALSE;
+		if(field<1) field=HDR_BODY;
 	}
 
 	switch (key->type)
@@ -342,6 +355,7 @@ static bool fts_backend_xapian_update_set_build_key(struct fts_backend_update_co
 			ctx->tbi_uid=key->uid;
 			break;
 		case FTS_BACKEND_BUILD_KEY_BODY_PART:
+			ctx->tbi_isfield=false;
 			ctx->tbi_uid=key->uid;
 			break;
 		case FTS_BACKEND_BUILD_KEY_BODY_PART_BINARY:
@@ -349,6 +363,8 @@ static bool fts_backend_xapian_update_set_build_key(struct fts_backend_update_co
 			return FALSE;
 	}
 
+	ctx->tbi_field = i_strdup_printf("%ld",field);
+
 	if((ctx->tbi_uid>0) && (ctx->tbi_uid != backend->lastuid))
         {
 		std::string s("FTS Xapian: New doc incoming (#");
@@ -356,47 +372,50 @@ static bool fts_backend_xapian_update_set_build_key(struct fts_backend_update_co
 
 		if(fts_xapian_settings.verbose>0) i_info("%s",s.c_str());
 
-		fts_backend_xapian_get_lock(backend, fts_xapian_settings.verbose, s.c_str());
-
 		if(backend->threads.size() < backend->max_threads )
                 {
-                        XDocsWriter * x = new XDocsWriter(backend,backend->threads.size()+1);
-                        x->launch(s.c_str());
+                       	XDocsWriter * x = new XDocsWriter(backend,backend->threads.size()+1);
+                       	x->launch(s.c_str());
 			backend->threads.push_back(x);
                 }
                        
                 long n = backend->threads.size();
-                while(n>0)
-                {
-                        n--;
-                        if(!(backend->threads.at(n)->started))
-                        {
-                                backend->threads[n]->launch("Relaunch post error");
-                        }
-                }
-
-		if(backend->lastuid>0)
-		{
-			if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Previous doc ready to index (#%ld)",backend->lastuid);
-			backend->docs.front()->status=1;	
-		}
-                backend->lastuid = ctx->tbi_uid;
-		backend->docs.insert(backend->docs.begin(),new XDoc(backend));
+               	while(n>0)
+               	{
+                       	n--;
+                       	if(!(backend->threads.at(n)->started))
+                       	{
+                       	        backend->threads[n]->launch("Relaunch post error");
+                       	}
+               	}
 		
-		if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Start indexing #%ld (%s) : Queue size = %ld",backend->lastuid, backend->boxname,backend->docs.size());
-
+		fts_backend_xapian_get_lock(backend, fts_xapian_settings.verbose, s.c_str());
+		{
+			if(backend->lastuid>0)
+			{
+				if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Previous doc ready to index (#%ld)",backend->lastuid);
+				backend->docs.front()->status=1;	
+			}
+                	backend->lastuid = ctx->tbi_uid;
+			backend->docs.insert(backend->docs.begin(),new XDoc(backend));
+		
+			if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Start indexing #%ld (%s) : Queue size = %ld",backend->lastuid, backend->boxname,backend->docs.size());
+		}
 		fts_backend_xapian_release_lock(backend, fts_xapian_settings.verbose, s.c_str());
 
-		n=0;
-		while(backend->docs.size()>3*backend->max_threads)
+		if(backend->docs.size() > (XAPIAN_WRITING_CACHE * 2) )
 		{
-			n++;
-			if((n>50) && (fts_xapian_settings.verbose>0))
+			n=0;
+			while (backend->docs.size() > XAPIAN_WRITING_CACHE )
 			{
-				i_info("FTS Xapian: Waiting for queue to be absorbed (pending : %ld vs limit :%ld)",backend->docs.size(),3*backend->max_threads);
-				n=0;
+				n++;
+				if(n>50)
+				{
+					if (fts_xapian_settings.verbose>0) i_info("FTS Xapian: Waiting for queue to be absorbed (pending=%ld)",backend->docs.size());
+					n=0;
+				}
+				std::this_thread::sleep_for(XAPIAN_SLEEP);
 			}
-			std::this_thread::sleep_for(XAPIAN_SLEEP);
 		}
         }
 
@@ -421,8 +440,6 @@ static int fts_backend_xapian_refresh(struct fts_backend * _backend)
 {
 	if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: fts_backend_xapian_refresh");
 
-	struct xapian_fts_backend *backend = (struct xapian_fts_backend *) _backend;
-
 	return 0;
 }
 
@@ -436,15 +453,9 @@ static int fts_backend_xapian_update_build_more(struct fts_backend_update_contex
 	if(data == NULL) return 0;
 
 	const char * d = (const char *) data;
-	if(strlen(d)<fts_xapian_settings.partial) return 0;
+	if(strlen(d)<(unsigned long)fts_xapian_settings.partial) return 0;
 
-	int i=0;
-        while((i<HDRS_NB-1) && (strcmp(ctx->tbi_field,hdrs_emails[i])!=0))
-        {
-                i++;
-        }
-        if(i>=HDRS_NB) i=HDRS_NB-1;
-        const char * h = hdrs_xapian[i];
+	long h = atol(ctx->tbi_field);
 
 	backend->docs.front()->raw_load(h,d,size,fts_xapian_settings.verbose,"fts_backend_xapian_index");
         	
